--
-- Resistor with Self-Heating
-- 
-- Author: Allan Ecker
-- Created: July 23, 2003
--
-------------------------------------------------------------------------------
-- Description:
-- Model the electrical and thermal characteristics of a resistor 
-- simultaneously
-------------------------------------------------------------------------------
package electrical_system is
        nature electrical is real across real through;
        terminal ground : electrical;
end electrical_system;
        
use electrical_system.all;

package thermal_system is
        nature thermal is real across real through;
        terminal thermal_ground : thermal;
end thermal_system;

use thermal_system.all;
                                  

                                                   
entity Rheat is
	generic ( 
	        alpha : real := 1.0e-3;  
                tnom : real := 300.0;
                R : real := 1.0e3   
	);
	port ( 
		terminal pos, neg : electrical;
		terminal t1, t2 : thermal
	);
end Rheat;

architecture arch1 of Rheat is
	quantity volt across pos to neg;
	quantity curr through pos to neg;
	quantity temp across HeatProduced through t1 to t2;
begin
  Thermal_equations: procedural is
  begin -- procedural statements                              
	curr := volt/(R*(1.0+alpha*(temp+tnom)/tnom)); --thermal resistor
	HeatProduced := -(volt*curr); --heat generated
  end procedural;
end architecture arch1;
