-- Auther: Bo Wan

library IEEE; 

package electrical_system is
        nature electrical is real across real through;
        terminal ground : electrical;
end electrical_system;

use electrical_system.all;

package mosdata is 
  alias undefined is real'low; 

  constant eps0 : real :=8.85418e-12; 
  constant Ni : real :=1.45e16; 
  constant Boltzmann : real :=1.380662e-23; 
  constant echarge: real :=1.6021892e-19; 
  constant epsSiO2 : real :=3.9*eps0; 
  constant epsSi : real :=11.7*eps0; 
  constant pi: real := 3.14159; 
end package mosdata; 

use mosdata.all; 

entity mymos3 is 
  
  generic( 
    -- instance parameters 
    --width  : real:=1.0E-4;  -- should be global constant DEFW! 
	n : real := 1.0;
    channel: real :=1.0;     -- +1 for NMOS, -1  for PMOS 
    -- model parameters 
    vt0    : real := undefined; 
    kp     : real := 2.0E-5; 
    gamma  : real := undefined; 
    ngate  : real := 1.0E12; 		-- p731, hspice manual
    phi    : real := undefined; 
    tox    : real :=  1.0E-7; 
    nsub   : real :=  0.0; 
    nss    : real :=  0.0; 
    nfs    : real :=  0.0; 
    tpg    : real :=  1.0; 
    xj     : real :=  0.0; 
    ld     : real :=  0.0; 
    u0     : real :=  600.0; 
    vmax   : real :=  0.0; 
    xqc    : real :=  1.0; 
    kf     : real :=  0.0; 
    af     : real :=  1.0; 
    fc     : real :=  0.5; 
    delta  : real :=  0.0; 
    theta  : real :=  0.0; 
    eta    : real :=  0.0; 
    kappa  : real :=  0.2; 
    Temperature : real :=300.15;  -- Should be global 
    INST_width  : real:=1.0E-4;  -- should be global constant DEFW! 
    INST_length : real:=1.0E-4  -- DEFL 
    ); 
  port ( terminal  drain,  gate,  source,  bulk  : electrical); 
end entity mymos3; 

architecture mos3 of mymos3 is 

  quantity Qcq, Qbq, Qgq : real; -- channel, bulk and gate charges 
  quantity Vdsq across drain to source; 
  quantity Vgsq across gate to source; 
  quantity Vbsq across bulk to source; 
  quantity Idq through drain; 
  quantity Igq through gate; 
  quantity Isq through source; 
  quantity Ibq through bulk; 

begin 
  MOSequations: procedural is 
     variable Qc,Qb,Qg: real; 
     variable cox,beta,vt,v_nsub,sigma,v_Phi,V_Gamma,v_nss,v_ngate,A,B,C,D,
	   Vfb,fshort,egfet,fermig,V_Gamma_tmp,mobdeg, 
           wp,wc,sqwpxj,vbulk,delv,vth,Vgstos, Vgst, Ueff,Tau,Vsat,Vpp,fdrain, 
           stfct,leff,xd,qnfscox,fn,dcrit,deltal,It,Ids,R,Vds,Vgs,Vbs, 
           forward : real; 
     variable kTQ : real;
     variable beta_new,beta_new2,cox_new,VdsTmp,VsatTmp1,VsatTmp2 : real;
  begin -- procedural statements 

    u0_new := u0/10000.0;  -- scale u0 to SI units 
    vt0_new := channel*vt0;		-- wb
    kTQ :=Boltzmann*Temperature/echarge;

     -- initialisation 
      if tox<=0.0 then 
        cox:=epsSiO2/0.0000001; 
      else 
        cox:=epsSiO2/tox; 
      end if; 

      if kp = 0.0 then 
        beta:=cox*u0_new; 
      else 
        beta:=kp; 
      end if; 

     v_nsub := nsub * 1000000.0;  -- scale nsub to SI units 


      if (phi = undefined) then 
        if (v_nsub > 0.0) then 
	  		if (0.1 > 2.0*kTQ*ln(v_nsub/Ni)) then
             v_Phi:=0.1;
	  		else
             v_Phi:=2.0*kTQ*ln(v_nsub/Ni); 
	  		end if;
        else 
          v_Phi:=0.6; 
        end if; 
      else 
        v_Phi:=phi; 
      end if; -- phi = undefined 
  

      if (gamma = undefined) then 
        if (v_nsub > 0.0) then 
          V_Gamma:=sqrt(2.0*epsSi*echarge*v_nsub)/cox; 
        else 
          V_Gamma:=0.0; 
        end if; 
      else 
        V_Gamma:=gamma; 
      end if; -- gamma = undefined 

      v_nss:=nss*10000.0;              -- Scale to SI 

      v_ngate:=ngate * 10000.0;           -- Scale to SI 

      if (vt0_new = undefined) then 
        egfet:=1.16-(7.02e-4*Temperature*Temperature)/(Temperature+1108.0); 
        if tpg=0.0 then 
          fermig:=0.05+egfet/2.0; 
        else 
          if v_ngate>0.0 then 
            fermig:=tpg*channel*kTQ*ln(v_ngate/Ni); 
          else 
            fermig:=tpg*channel*egfet/2.0; 
          end if; 
        end if; 
        vt:=-fermig+channel*(v_Phi*0.5+V_Gamma*sqrt(v_Phi))-v_nss*echarge/cox; 
      else 
        vt:=vt0_new; 
      end if; -- vt0 = undefined 

      leff:=INST_length-2.0*ld; 
      if leff>0.0 then 
        Sigma:=eta*8.15e-22/(cox*leff*leff*leff); 
      else 
        Sigma:=0.0; 
      end if; -- leff>0 

      if v_nsub>0.0 then -- N.B. v_nsub was scaled, above. 
        xd:=sqrt(2.0*epsSi/(echarge*v_nsub)); 
      else 
        xd:=0.0; 
      end if; -- v_nsub >0 

      if (nfs>0.0) and(cox>0.0) then 
        qnfscox:=echarge*nfs/cox; 
      else 
        qnfscox:=0.0; 
      end if; --nfs > 0 

      if cox>0.0 then 
        fn:=delta*pi*epsSi*0.5/(cox*n); 
      else 
        fn:=delta*pi*epsSi*0.5*tox/epsSiO2; 
      end if;  -- cox > 0 

      beta_new:=beta*INST_width/leff; 
      cox_new:=cox*INST_width*leff; 

    VdsTmp:=channel*Vdsq; 
    if VdsTmp>=0.0 then 
		Vds:=VdsTmp + 0.0;
        Vgs:=channel* Vgsq; 
        Vbs:=channel* Vbsq; 
        forward:=1.0; 
    else 
        Vds:=-VdsTmp; 
        Vgs:=channel* Vgsq; 
        Vbs:=channel* Vbsq; 
        forward:=-1.0; 
    end if; -- Vds >=0 

    if Vbs<=0.0 then 
        A:=v_Phi-Vbs; 
        D:=sqrt(A); 
    else 
        D:=2.0*sqrt(v_Phi)*v_Phi/(2.0*v_Phi+Vbs); 
        A:=D**2; 
    end if; -- Vbs <= 0 

    Vfb:=vt-V_Gamma*sqrt(v_Phi)-Sigma*Vds; 
    if (xd=0.0) OR (xj=0.0) then 
        fshort:=1.0; 
    else 
        wp:=xd*D; 
        wc:=0.0631353*xj+0.8013292*wp-0.01110777*wp*wp/xj; 
        sqwpxj:=sqrt(1.0-(wp*wp/((wp+xj)*(wp+xj)))); 
        fshort:=1.0-((ld+wc)*sqwpxj-ld)/leff; 
    end if; -- xd or xj = 0 

    vbulk:=V_Gamma*fshort*D+fn*A; 
    if nfs=0.0 then 
        delv:=0.0; 
    else 
        delv:=kTQ*(1.0+qnfscox+vbulk*0.5/A); 
    end if; -- nfs = 0 

    vth:=Vfb+vbulk; 
    Vgstos:=Vgs-Vfb; 

    if (Vgs-vth > delv) then
       Vgst:=Vgs-vth;
    else
       Vgst:=delv;
    end if;

    if (Vgs>=vth) or (delv/=0.0) then 

      if (Vbs<=0.0) or (v_Phi /= 0.0) then 
          B:=0.25*V_Gamma/D+fn; 
      else 
          B:=fn; 
      end if; 

      mobdeg:=1.0/(1.0+theta*Vgst); 
      if (vmax /=0.0) then 
         Ueff:=u0_new*mobdeg; 
          Tau:=Ueff/(leff*vmax); 
      else 
	  	  Ueff := 0.0;
          Tau:=0.0; 
      end if; 

      Vsat:=Vgst/(1.0+B); 
      VsatTmp1:=1.0-0.5*Tau*Vsat;
      VsatTmp2:=Vsat*VsatTmp1;

      if (Vds < VsatTmp2) then
	 	Vpp:=Vds;
      else
	 	Vpp:=VsatTmp2;
      end if;

      fdrain:=1.0/(1.0+Tau*Vpp); 
      if (Vgs<vth+delv) and (nfs>0.0) then 
        stfct:=exp((Vgs-vth-delv)/delv); 
      else 
          stfct:=1.0; 
      end if; 

      if Vds>=VsatTmp2 then 
        if (kappa>0.0) and (xd>0.0) then 

          if vmax=0.0 then 
              deltal:=sqrt(kappa*xd*xd*(Vds-VsatTmp2)); 
          else 
              dcrit:=((xd**2) *vmax*0.5)/(Ueff*(1.0-fdrain)); 
              deltal:=sqrt(kappa*(xd**2)*(Vds-VsatTmp2)+(dcrit**2))-dcrit; 
          end if; 

          if deltal<=0.5*leff then 
              C:=leff/(leff-deltal); 
          else 
              C:=4.0*deltal/leff; 
          end if; 

        else --kappa=0.0 or xd=0.0 
            C:=1.0; 
        end if; 

      else 
          C:=1.0; 
      end if; 

      It:=Vgst-Vpp*(1.0+B)*0.5; 
        beta_new2:=beta_new*mobdeg; 
        Ids:=beta_new2*Vpp*It*C*fdrain*stfct; 
    else 
        -- Cutoff 
		B:=0.0;
		Vpp:=0.0;
		It:=0.0;
        Ids:=0.0; 
    end if; -- Vgs >= vth 

    if cox_new /= 0.0 then 
        --Charges 
      if Vgs<=vth then 
        if V_Gamma /= 0.0 then 
          if Vgstos < -A then 
           Qg:=cox_new*(Vgstos+A); -- Accumulation 
          else 
           Qg:=0.5*V_Gamma*cox_new*(sqrt(4.0*(Vgstos+A)+sqrt(V_Gamma))-V_Gamma);
          end if ; -- Vgstos <-A 
        else-- V_Gamma = 0.0 
            Qg:=0.0; 
        end if; -- gamma /= 0 
          Qb:=-Qg; 
          Qc:=0.0; 
      else 
          -- depletion mode: 
          R:=(1.0+B)*Vpp*Vpp/(12.0*It); 
          Qg:=cox_new*(Vgstos-Vpp*0.5+R); 
          Qc:=-cox_new*(Vgst+(1.0+B)*(R-Vpp*0.5)); 
          Qb:=-(Qc+Qg); 
      end if; -- Vgs<=vth 

    else 
        Qg:=0.0; 
        Qc:=0.0; 
        Qb:=0.0; 
    end if; -- cox_new /= 0 

    -- equations for charges (in a procedural we have assignments to quantitites): 
      Qcq := Qc; 
      Qgq := Qg; 
      Qbq := Qb; 

    -- equations for currents: 
      Idq := channel*forward*Ids+channel*xqc*Qc'dot;
      Igq := channel*Qg'dot;
      Ibq := channel*Qb'dot;
      Isq := -Idq - Igq - Ibq; 

  end procedural; 
end architecture mos3; 
  
