/****************************************************************
File generated by Device Model Compiler: BMAC
Author: Bo Wan, 2003
Copyright 2003(c) Mixed-Signal CAD lab. University of Washington.
All rights reserved.
****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/timeb.h>
#include "genCommon.h"

extern double **dmatrix(int,int,int,int);
extern void free_dmatrix(double**,long,long,long,long);
extern double uniform_interpolate_bilinear(double**,double**,double***,int,int,double,double);
extern double log_interpolate_bilinear(TableList*, double, double);
//extern double interpolate_bilinear(double*,double*,double**,int,int,double,double);
//extern double interpolate_eslq3(double*,double*,double**,int,int,double,double);
//extern double interpolate_eslgq(double*,double*,double**,int,int,double,double);

static const int BSIZE = 4096;
static TableList* table_list = NULL;

static struct timeb timebegin;

void startTimer() {ftime(&timebegin);}

void timediff(struct timeb* now, struct timeb* begin, int* sec, int* msec) {
  *msec = now->millitm - begin->millitm;
  *sec = now->time - begin->time;
  if (*msec < 0) {
    *msec += 1000;
    (*sec)--;
  }
  return;
}

double runInSeconds() {
  struct timeb timenow;
  int sec, msec;

  ftime(&timenow);
  timediff(&timenow, &timebegin, &sec, &msec);
  return (sec + (double) msec / 1000.0);
}

void freeTables() {
  TableList* header = table_list;
  while (header!=NULL) {
	free_dmatrix(header->_table,0,header->_m-1,0,header->_n-1);
	header = header->_next;
  }
}

void readTable(TableList* current) {
  char* filename;
  FILE* tableFile;
  int m,n;
  char* token;
  int i,j;
  char buf[BSIZE];
  double** table;

  filename = current->_name;
  if (!(tableFile = fopen(filename, "r"))) {
	fprintf(stderr, "Can not open Table data file \"%s\" !\n",filename);
	exit(1);
  }

  // read the dimension m x n
  while (fgets(buf, BSIZE, tableFile)) {
    if (!strncmp("#", buf, 1)) continue;
    //if (sscanf(buf, "%d %d xLog=%d yLog=%d xLogStep=%-.15g yLogStep=%-.15g", &m, &n, &_x_logFlag, &_y_logFlag, &_x_logStep, &_y_logStep) != 6) {
    if (sscanf(buf, "%d %d xLog=%d yLog=%d xLogStep=%lf yLogStep=%lf", &m, &n, &(current->_x_logFlag), &(current->_y_logFlag), &(current->_x_logStep), &(current->_y_logStep)) != 6) {
    	fprintf(stderr, "Error reading table data: %s\n", buf);
        exit(1);
    }
    break;
  }

  // allocate the arraies of x & y
  current->_x = (double*) malloc (sizeof(double)*m);
  current->_y = (double*) malloc (sizeof(double)*n);
  // x
  i=0;
  fgets(buf, BSIZE, tableFile);
  token = strtok(buf," ");
  current->_x[i] = atof(token);
  for (i=1;i<m;i++) {
	token = strtok(NULL, " ");
	current->_x[i] = atof(token);
  }
  // y
  j=0;
  fgets(buf, BSIZE, tableFile);
  token = strtok(buf," ");
  current->_y[j] = atof(token);
  for (j=1;j<n;j++) {
	token = strtok(NULL, " ");
	current->_y[j] = atof(token);
  }
  
  // allocate table matrix
  table = dmatrix(0,m-1,0,n-1);

  // fill in the matrix from read data
  i=0;
  while (fgets(buf, BSIZE, tableFile)) {
	if (!strncmp("#", buf, 1)) continue;
	j=0;
	token = strtok(buf, " ");
	//printf("token: %s\n",token);
	table[i][j] = atof(token);
	for (j=1;j<n;j++) {
		token = strtok(NULL, " ");
		table[i][j] = atof(token);
	}
	i++;
  }
  //close file handle
  (void) fclose(tableFile);

  //sanity checking
  if (i != m) {
	fprintf(stderr, "Error format !\n");
	exit(1);
  }

  //fill in data
  current->_table = table;
  current->_m = m;
  current->_n = n;
}

int readTableIndex() {
  FILE* index;
  char buf[BSIZE], tableName[BSIZE];
  TableList* tmp;

  if (!(index = fopen("TableIndex", "r"))) {
	fprintf(stderr, "Warning: Can not find file named \"TableIndex\" \n");
	fprintf(stderr, "\tif NO table(s) generated, just ignore this warning! \n");
	fprintf(stderr, "\tBUT if DO generate table(s), this is an ERROR!!! \n\n");
	return 1;
	//exit(1);
  }
  while (fgets(buf, BSIZE, index)) {
	if (!strncmp("#", buf, 1)) continue;
	if (sscanf(buf, "%s", &tableName) != 1) {
		fprintf(stderr, "Error reading index data: %s\n", buf);
		exit(1);
	}
    //printf("TableName: %s\n", tableName);

  	// add in the front of the table_list
  	tmp = (TableList*) malloc(sizeof(TableList));
  	strcpy(tmp->_name,tableName);
  	//strncpy(tmp->_name,tableName,strlen(tableName)-3); //remove "_TB"
  	tmp->_next = table_list;
  	table_list = tmp;
  }
  (void) fclose(index);
  return 0;
}

// read all tables from table files and put them in the TableList
void loadTables() {
  TableList* header;

  if (readTableIndex()) return;

  header = table_list;
  while (header) {
    readTable(header);
	header = header->_next;
  }
}

double tableLookup(char* name, double u, double v) {
  TableList* curr;
  double result;
  char tableName[256];

  // find table first
  strcpy(tableName,name);
  strcat(tableName,"_TB");
  curr = table_list;
  while (curr) {
	if (!strcmp(curr->_name,tableName)) break;
	curr = curr->_next;
  }
  if (!curr) {
	fprintf(stderr, "Can not find table file: %s_TB !\n", tableName);
	exit(1);
  }

  if (curr->_x_logFlag || curr->_y_logFlag) {
  	result = log_interpolate_bilinear(curr,u,v);
  } else {
  	//result = interpolate_bilinear(curr->_x, curr->_y,
  	result = uniform_interpolate_bilinear(&(curr->_x), &(curr->_y),
				  &(curr->_table),
				  curr->_m,curr->_n,
				  u,v);
  }

  return result;
}

double cc_evaluate() {
//  #include "cc_optimized.gc"
}

